@@ -1,4 +1,4 @@
-# COPYRIGHT (C) 2021-2022 Nicotine+ Team
+# COPYRIGHT (C) 2021-2022 Nicotine+ Contributors
 #
 # GNU GENERAL PUBLIC LICENSE
 #    Version 3, 29 June 2007
@@ -17,10 +17,24 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 import os
+import sys
 
 
 def check_gui_dependencies():
 
+    # Defaults for different operating systems
+    components = {
+        "gtk": {"win32": '3', "darwin": '3'},
+        "libadwaita": {"win32": '0', "darwin": '0'}
+    }
+
+    if os.getenv("NICOTINE_GTK_VERSION") is None:
+        os.environ["NICOTINE_GTK_VERSION"] = components.get("gtk").get(sys.platform, '3')
+
+    if os.getenv("NICOTINE_LIBADWAITA") is None:
+        os.environ["NICOTINE_LIBADWAITA"] = components.get("libadwaita").get(sys.platform, '0')
+
+    # Require minor version of GTK
     if os.getenv("NICOTINE_GTK_VERSION") == '4':
         gtk_version = (4, 6, 0)
         pygobject_version = (3, 40, 0)
@@ -66,9 +80,22 @@ def check_gui_dependencies():
     return None
 
 
-def run_gui(network_processor, trayicon, hidden, bindip, port, ci_mode, multi_instance):
+def run_gui(core, trayicon, hidden, bindip, port, ci_mode, multi_instance):
     """ Run Nicotine+ GTK GUI """
 
+    if getattr(sys, 'frozen', False):
+        # Set up paths for frozen binaries (Windows and macOS)
+        executable_folder = os.path.dirname(sys.executable)
+        resources_folder = executable_folder
+
+        if sys.platform == "darwin":
+            resources_folder = os.path.abspath(os.path.join(executable_folder, "..", "Resources"))
+
+        os.environ["XDG_DATA_DIRS"] = os.path.join(resources_folder, "share")
+        os.environ["GDK_PIXBUF_MODULE_FILE"] = os.path.join(executable_folder, "lib/pixbuf-loaders.cache")
+        os.environ["GI_TYPELIB_PATH"] = os.path.join(executable_folder, "lib/typelibs")
+        os.environ["GSETTINGS_SCHEMA_DIR"] = os.path.join(executable_folder, "lib/schemas")
+
     from pynicotine.logfacility import log
     error = check_gui_dependencies()
 
@@ -76,5 +103,5 @@ def run_gui(network_processor, trayicon, hidden, bindip, port, ci_mode, multi_in
         log.add(error)
         return 1
 
-    from pynicotine.gtkgui.frame import Application
-    return Application(network_processor, trayicon, hidden, bindip, port, ci_mode, multi_instance).run()
+    from pynicotine.gtkgui.application import Application
+    return Application(core, trayicon, hidden, bindip, port, ci_mode, multi_instance).run()
