@@ -1,4 +1,4 @@
-# COPYRIGHT (C) 2020-2022 Nicotine+ Team
+# COPYRIGHT (C) 2020-2022 Nicotine+ Contributors
 # COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
 # COPYRIGHT (C) 2008-2011 Quinox <quinox@users.sf.net>
 # COPYRIGHT (C) 2007 Gallows <g4ll0ws@gmail.com>
@@ -27,13 +27,14 @@ from collections import deque
 
 from gi.repository import Gio
 from gi.repository import GLib
-from gi.repository import Gtk
 
 from pynicotine import slskmessages
 from pynicotine.config import config
+from pynicotine.gtkgui.application import GTK_API_VERSION
 from pynicotine.gtkgui.widgets.iconnotebook import IconNotebook
 from pynicotine.gtkgui.widgets.popupmenu import PopupMenu
-from pynicotine.gtkgui.widgets.dialogs import option_dialog
+from pynicotine.gtkgui.widgets.popupmenu import UserPopupMenu
+from pynicotine.gtkgui.widgets.dialogs import OptionDialog
 from pynicotine.gtkgui.widgets.textentry import ChatCompletion
 from pynicotine.gtkgui.widgets.textentry import ChatEntry
 from pynicotine.gtkgui.widgets.textentry import TextSearchBar
@@ -42,58 +43,76 @@ from pynicotine.gtkgui.widgets.theme import get_user_status_color
 from pynicotine.gtkgui.widgets.theme import update_widget_visuals
 from pynicotine.gtkgui.widgets.ui import UserInterface
 from pynicotine.logfacility import log
+from pynicotine.slskmessages import UserStatus
 from pynicotine.utils import clean_file
 from pynicotine.utils import delete_log
+from pynicotine.utils import encode_path
 from pynicotine.utils import open_log
 
 
 class PrivateChats(IconNotebook):
 
-    def __init__(self, frame):
+    def __init__(self, frame, core):
 
-        IconNotebook.__init__(self, frame, frame.private_notebook, "private")
-        self.notebook.connect("switch-page", self.on_switch_chat)
+        super().__init__(
+            frame, core,
+            widget=frame.private_notebook,
+            parent_page=frame.private_page,
+            switch_page_callback=self.on_switch_chat
+        )
 
         self.completion = ChatCompletion()
-        self.command_help = UserInterface("ui/popovers/privatechatcommands.ui")
 
-        if Gtk.get_major_version() == 4:
+        self.command_help = UserInterface("ui/popovers/privatechatcommands.ui")
+        self.command_help.container, self.command_help.popover = self.command_help.widgets
+
+        if GTK_API_VERSION >= 4:
             # Scroll to the focused widget
             self.command_help.container.get_child().set_scroll_to_focus(True)
 
     def on_switch_chat(self, _notebook, page, _page_num):
 
-        if self.frame.current_page_id != self.page_id:
+        if self.frame.current_page_id != self.frame.private_page.id:
             return
 
         for user, tab in self.pages.items():
-            if tab.Main == page:
-                GLib.idle_add(lambda: tab.ChatLine.grab_focus() == -1)  # pylint:disable=cell-var-from-loop
+            if tab.container != page:
+                continue
 
-                self.completion.set_entry(tab.ChatLine)
-                tab.set_completion_list(list(self.frame.np.privatechats.completion_list))
+            GLib.idle_add(lambda tab: tab.chat_entry.grab_focus() == -1, tab)
 
-                self.command_help.popover.unparent()
-                tab.ShowChatHelp.set_popover(self.command_help.popover)
+            self.completion.set_entry(tab.chat_entry)
+            tab.set_completion_list(self.core.privatechats.completion_list[:])
 
-                # If the tab hasn't been opened previously, scroll chat to bottom
-                if not tab.opened:
-                    GLib.idle_add(tab.chat_textview.scroll_bottom)
-                    tab.opened = True
+            self.command_help.popover.unparent()
+            tab.help_button.set_popover(self.command_help.popover)
 
-                # Remove hilite if selected tab belongs to a user in the hilite list
-                self.frame.notifications.clear("private", user)
-                break
+            if not tab.loaded:
+                tab.load()
+
+            # Remove hilite if selected tab belongs to a user in the hilite list
+            self.frame.notifications.clear("private", user)
+            break
+
+    def on_get_private_chat(self, *_args):
+
+        username = self.frame.private_entry.get_text().strip()
+
+        if not username:
+            return
+
+        self.frame.private_entry.set_text("")
+        self.core.privatechats.show_user(username)
 
     def clear_notifications(self):
 
-        if self.frame.current_page_id != self.page_id:
+        if self.frame.current_page_id != self.frame.private_page.id:
             return
 
-        page = self.get_nth_page(self.get_current_page())
+        page = self.get_current_page()
 
         for user, tab in self.pages.items():
-            if tab.Main == page:
+            if tab.container == page:
                 # Remove hilite
                 self.frame.notifications.clear("private", user)
                 break
@@ -102,10 +121,10 @@ class PrivateChats(IconNotebook):
 
         page = self.pages.get(msg.user)
         if page is not None:
-            self.set_user_status(page.Main, msg.user, msg.status)
+            self.set_user_status(page.container, msg.user, msg.status)
             page.update_remote_username_tag(msg.status)
 
-        if msg.user == self.frame.np.login_username:
+        if msg.user == self.core.login_username:
             for page in self.pages.values():
                 # We've enabled/disabled away mode, update our username color in all chats
                 page.update_local_username_tag(msg.status)
@@ -114,11 +133,22 @@ class PrivateChats(IconNotebook):
 
         if user not in self.pages:
             self.pages[user] = page = PrivateChat(self, user)
-            self.append_page(page.Main, user, page.on_close, user=user)
-            page.set_label(self.get_tab_label_inner(page.Main))
+            self.append_page(page.container, user, page.on_close, user=user)
+            page.set_label(self.get_tab_label_inner(page.container))
 
-        if switch_page and self.get_current_page() != self.page_num(self.pages[user].Main):
-            self.set_current_page(self.page_num(self.pages[user].Main))
+        if switch_page and self.get_current_page() != self.pages[user].container:
+            self.set_current_page(self.pages[user].container)
+
+    def remove_user(self, user):
+
+        page = self.pages.get(user)
+
+        if page is None:
+            return
+
+        page.clear()
+        self.remove_page(page.container)
+        del self.pages[user]
 
     def echo_message(self, user, text, message_type):
 
@@ -138,13 +168,17 @@ class PrivateChats(IconNotebook):
         if page is not None:
             page.message_user(msg)
 
+    def toggle_chat_buttons(self):
+        for page in self.pages.values():
+            page.toggle_chat_buttons()
+
     def set_completion_list(self, completion_list):
 
-        page = self.get_nth_page(self.get_current_page())
+        page = self.get_current_page()
 
         for tab in self.pages.values():
-            if tab.Main == page:
-                tab.set_completion_list(list(completion_list))
+            if tab.container == page:
+                tab.set_completion_list(completion_list[:])
                 break
 
     def update_visuals(self):
@@ -161,7 +195,7 @@ class PrivateChats(IconNotebook):
 
         for user, page in self.pages.items():
             page.server_disconnect()
-            self.set_user_status(page.Main, user, 0)
+            self.set_user_status(page.container, user, UserStatus.OFFLINE)
 
 
 class PrivateChat(UserInterface):
@@ -169,32 +203,42 @@ class PrivateChat(UserInterface):
     def __init__(self, chats, user):
 
         super().__init__("ui/privatechat.ui")
+        (
+            self.chat_entry,
+            self.chat_view,
+            self.container,
+            self.help_button,
+            self.log_toggle,
+            self.search_bar,
+            self.search_entry,
+            self.speech_toggle
+        ) = self.widgets
 
         self.user = user
         self.chats = chats
         self.frame = chats.frame
+        self.core = chats.core
 
-        self.opened = False
+        self.loaded = False
         self.offline_message = False
-        self.status = 0
+        self.status = self.core.user_statuses.get(user, UserStatus.OFFLINE)
 
-        if user in self.frame.np.user_statuses:
-            self.status = self.frame.np.user_statuses[user] or 0
+        self.chat_view = TextView(self.chat_view, font="chatfont")
 
         # Text Search
-        TextSearchBar(self.ChatScroll, self.SearchBar, self.SearchEntry,
-                      controller_widget=self.Main, focus_widget=self.ChatLine)
-
-        self.chat_textview = TextView(self.ChatScroll, font="chatfont")
+        self.search_bar = TextSearchBar(self.chat_view.textview, self.search_bar, self.search_entry,
+                                        controller_widget=self.container, focus_widget=self.chat_entry)
 
         # Chat Entry
-        ChatEntry(self.frame, self.ChatLine, chats.completion, user, slskmessages.MessageUser,
-                  self.frame.np.privatechats.send_message, self.frame.np.privatechats.CMDS)
+        ChatEntry(self.frame, self.chat_entry, chats.completion, user, slskmessages.MessageUser,
+                  self.core.privatechats.send_message, self.core.privatechats.CMDS)
 
-        self.Log.set_active(config.sections["logging"]["privatechat"])
+        self.log_toggle.set_active(config.sections["logging"]["privatechat"])
 
-        self.popup_menu_user_chat = PopupMenu(self.frame, self.ChatScroll, connect_events=False)
-        self.popup_menu_user_tab = PopupMenu(self.frame, None, self.on_popup_menu_user)
+        self.toggle_chat_buttons()
+
+        self.popup_menu_user_chat = UserPopupMenu(self.frame, self.chat_view.textview, connect_events=False)
+        self.popup_menu_user_tab = UserPopupMenu(self.frame, None, self.on_popup_menu_user)
 
         for menu in (self.popup_menu_user_chat, self.popup_menu_user_tab):
             menu.setup_user_menu(user, page="privatechat")
@@ -204,18 +248,18 @@ class PrivateChat(UserInterface):
                 ("#" + _("_Close Tab"), self.on_close)
             )
 
-        popup = PopupMenu(self.frame, self.ChatScroll, self.on_popup_menu_chat)
+        popup = PopupMenu(self.frame, self.chat_view.textview, self.on_popup_menu_chat)
         popup.add_items(
             ("#" + _("Find…"), self.on_find_chat_log),
             ("", None),
-            ("#" + _("Copy"), self.chat_textview.on_copy_text),
-            ("#" + _("Copy Link"), self.chat_textview.on_copy_link),
-            ("#" + _("Copy All"), self.chat_textview.on_copy_all_text),
+            ("#" + _("Copy"), self.chat_view.on_copy_text),
+            ("#" + _("Copy Link"), self.chat_view.on_copy_link),
+            ("#" + _("Copy All"), self.chat_view.on_copy_all_text),
             ("", None),
             ("#" + _("View Chat Log"), self.on_view_chat_log),
             ("#" + _("Delete Chat Log…"), self.on_delete_chat_log),
             ("", None),
-            ("#" + _("Clear Message View"), self.chat_textview.on_clear_all_text),
+            ("#" + _("Clear Message View"), self.chat_view.on_clear_all_text),
             ("", None),
             (">" + _("User"), self.popup_menu_user_tab),
         )
@@ -225,6 +269,12 @@ class PrivateChat(UserInterface):
 
         self.read_private_log()
 
+    def load(self):
+
+        # Scroll chat to bottom
+        GLib.idle_add(self.chat_view.scroll_bottom)
+        self.loaded = self.chat_view.auto_scroll = True
+
     def read_private_log(self):
 
         # Read log file
@@ -239,7 +289,7 @@ class PrivateChat(UserInterface):
 
     def append_log_lines(self, path, numlines):
 
-        with open(path, "rb") as lines:
+        with open(encode_path(path), "rb") as lines:
             # Only show as many log lines as specified in config
             lines = deque(lines, numlines)
 
@@ -250,22 +300,24 @@ class PrivateChat(UserInterface):
                 except UnicodeDecodeError:
                     line = line.decode("latin-1")
 
-                self.chat_textview.append_line(line, self.tag_hilite, timestamp_format="", scroll=False)
+                self.chat_view.append_line(line, tag=self.tag_hilite)
 
     def server_login(self):
         timestamp_format = config.sections["logging"]["private_timestamp"]
-        self.chat_textview.append_line(_("--- reconnected ---"), self.tag_hilite, timestamp_format=timestamp_format)
+        self.chat_view.append_line(_("--- reconnected ---"), tag=self.tag_hilite, timestamp_format=timestamp_format)
 
     def server_disconnect(self):
 
         timestamp_format = config.sections["logging"]["private_timestamp"]
-        self.chat_textview.append_line(_("--- disconnected ---"), self.tag_hilite, timestamp_format=timestamp_format)
-        self.status = -1
+        self.chat_view.append_line(_("--- disconnected ---"), tag=self.tag_hilite, timestamp_format=timestamp_format)
         self.offline_message = False
 
-        # Offline color for usernames
-        self.update_remote_username_tag(status=0)
-        self.update_local_username_tag(status=0)
+        self.update_remote_username_tag(status=UserStatus.OFFLINE)
+        self.update_local_username_tag(status=UserStatus.OFFLINE)
+
+    def clear(self):
+        self.chat_view.clear()
+        self.frame.notifications.clear("private", self.user)
 
     def set_label(self, label):
         self.popup_menu_user_tab.set_parent(label)
@@ -274,41 +326,43 @@ class PrivateChat(UserInterface):
 
         self.popup_menu_user_tab.toggle_user_items()
 
-        menu.actions[_("Copy")].set_enabled(self.chat_textview.get_has_selection())
-        menu.actions[_("Copy Link")].set_enabled(bool(self.chat_textview.get_url_for_selected_pos()))
+        menu.actions[_("Copy")].set_enabled(self.chat_view.get_has_selection())
+        menu.actions[_("Copy Link")].set_enabled(bool(self.chat_view.get_url_for_current_pos()))
 
     def on_popup_menu_user(self, _menu, _widget):
         self.popup_menu_user_tab.toggle_user_items()
 
+    def toggle_chat_buttons(self):
+        self.speech_toggle.set_visible(config.sections["ui"]["speechenabled"])
+
     def on_find_chat_log(self, *_args):
-        self.SearchBar.set_search_mode(True)
+        self.search_bar.show()
 
     def on_view_chat_log(self, *_args):
         open_log(config.sections["logging"]["privatelogsdir"], self.user)
 
-    def on_delete_chat_log_response(self, dialog, response_id, _data):
-
-        dialog.destroy()
+    def on_delete_chat_log_response(self, _dialog, response_id, _data):
 
         if response_id == 2:
             delete_log(config.sections["logging"]["privatelogsdir"], self.user)
-            self.chat_textview.clear()
+            self.chats.history.remove_user(self.user)
+            self.chat_view.clear()
 
     def on_delete_chat_log(self, *_args):
 
-        option_dialog(
-            parent=self.frame.MainWindow,
+        OptionDialog(
+            parent=self.frame.window,
             title=_('Delete Logged Messages?'),
             message=_('Do you really want to permanently delete all logged messages for this user?'),
             callback=self.on_delete_chat_log_response
-        )
+        ).show()
 
     def show_notification(self, text):
 
-        self.chats.request_tab_hilite(self.Main)
+        self.chats.request_tab_hilite(self.container)
 
-        if (self.chats.get_current_page() == self.chats.page_num(self.Main)
-                and self.frame.current_page_id == self.chats.page_id and self.frame.MainWindow.is_active()):
+        if (self.chats.get_current_page() == self.container
+                and self.frame.current_page_id == self.frame.private_page.id and self.frame.window.is_active()):
             # Don't show notifications if the chat is open and the window is in use
             return
 
@@ -334,9 +388,11 @@ class PrivateChat(UserInterface):
         if text.startswith("/me "):
             line = "* %s %s" % (self.user, text[4:])
             tag = self.tag_action
+            speech = line[2:]
         else:
             line = "[%s] %s" % (self.user, text)
             tag = self.tag_remote
+            speech = text
 
         timestamp_format = config.sections["logging"]["private_timestamp"]
 
@@ -344,19 +400,26 @@ class PrivateChat(UserInterface):
             tag = usertag = self.tag_hilite
 
             if not self.offline_message:
-                self.chat_textview.append_line(_("* Message(s) sent while you were offline. Timestamps are reported by "
-                                                 "the server and can be off."), tag, timestamp_format=timestamp_format)
+                self.chat_view.append_line(_("* Message(s) sent while you were offline."), tag=tag,
+                                           timestamp_format=timestamp_format)
                 self.offline_message = True
 
         else:
             self.offline_message = False
 
-        self.chat_textview.append_line(line, tag, timestamp=timestamp, timestamp_format=timestamp_format,
-                                       username=self.user, usertag=usertag)
+        self.chat_view.append_line(line, tag=tag, timestamp=timestamp, timestamp_format=timestamp_format,
+                                   username=self.user, usertag=usertag)
 
-        if self.Log.get_active():
-            log.write_log(config.sections["logging"]["privatelogsdir"], self.user, line,
-                          timestamp, timestamp_format=config.sections["logging"]["log_timestamp"])
+        if self.speech_toggle.get_active():
+            self.core.notifications.new_tts(
+                config.sections["ui"]["speechprivate"], {"user": self.user, "message": speech}
+            )
+
+        if self.log_toggle.get_active():
+            log.write_log_file(
+                folder_path=config.sections["logging"]["privatelogsdir"], base_name=clean_file(self.user) + ".log",
+                text=line, timestamp=timestamp
+            )
 
     def echo_message(self, text, message_type):
 
@@ -366,11 +429,11 @@ class PrivateChat(UserInterface):
         if hasattr(self, "tag_" + str(message_type)):
             tag = getattr(self, "tag_" + str(message_type))
 
-        self.chat_textview.append_line(text, tag, timestamp_format=timestamp_format)
+        self.chat_view.append_line(text, tag=tag, timestamp_format=timestamp_format)
 
     def send_message(self, text):
 
-        my_username = self.frame.np.login_username
+        my_username = self.core.login_username
 
         if text.startswith("/me "):
             line = "* %s %s" % (my_username, text[4:])
@@ -379,16 +442,17 @@ class PrivateChat(UserInterface):
             line = "[%s] %s" % (my_username, text)
             tag = self.tag_local
 
-        self.chat_textview.append_line(line, tag, timestamp_format=config.sections["logging"]["private_timestamp"],
-                                       username=my_username, usertag=self.tag_my_username)
+        self.chat_view.append_line(line, tag=tag, timestamp_format=config.sections["logging"]["private_timestamp"],
+                                   username=my_username, usertag=self.tag_my_username)
 
-        if self.Log.get_active():
-            log.write_log(config.sections["logging"]["privatelogsdir"], self.user, line,
-                          timestamp_format=config.sections["logging"]["log_timestamp"])
+        if self.log_toggle.get_active():
+            log.write_log_file(
+                folder_path=config.sections["logging"]["privatelogsdir"],
+                base_name=clean_file(self.user) + ".log", text=line
+            )
 
     def update_visuals(self):
-
-        for widget in list(self.__dict__.values()):
+        for widget in self.__dict__.values():
             update_widget_visuals(widget)
 
     def user_name_event(self, pos_x, pos_y, user):
@@ -400,21 +464,21 @@ class PrivateChat(UserInterface):
 
     def create_tags(self):
 
-        self.tag_remote = self.chat_textview.create_tag("chatremote")
-        self.tag_local = self.chat_textview.create_tag("chatlocal")
-        self.tag_action = self.chat_textview.create_tag("chatme")
-        self.tag_hilite = self.chat_textview.create_tag("chathilite")
+        self.tag_remote = self.chat_view.create_tag("chatremote")
+        self.tag_local = self.chat_view.create_tag("chatlocal")
+        self.tag_action = self.chat_view.create_tag("chatme")
+        self.tag_hilite = self.chat_view.create_tag("chathilite")
 
         color = get_user_status_color(self.status)
-        self.tag_username = self.chat_textview.create_tag(color, callback=self.user_name_event, username=self.user)
+        self.tag_username = self.chat_view.create_tag(color, callback=self.user_name_event, username=self.user)
 
-        if not self.frame.np.logged_in:
+        if not self.core.logged_in:
             color = "useroffline"
         else:
-            color = "useraway" if self.frame.np.away else "useronline"
+            color = "useraway" if self.core.away else "useronline"
 
         my_username = config.sections["server"]["login"]
-        self.tag_my_username = self.chat_textview.create_tag(color, callback=self.user_name_event, username=my_username)
+        self.tag_my_username = self.chat_view.create_tag(color, callback=self.user_name_event, username=my_username)
 
     def update_remote_username_tag(self, status):
 
@@ -424,25 +488,20 @@ class PrivateChat(UserInterface):
         self.status = status
 
         color = get_user_status_color(status)
-        self.chat_textview.update_tag(self.tag_username, color)
+        self.chat_view.update_tag(self.tag_username, color)
 
     def update_local_username_tag(self, status):
         color = get_user_status_color(status)
-        self.chat_textview.update_tag(self.tag_my_username, color)
+        self.chat_view.update_tag(self.tag_my_username, color)
 
     def update_tags(self):
 
         for tag in (self.tag_remote, self.tag_local, self.tag_action, self.tag_hilite,
                     self.tag_username, self.tag_my_username):
-            self.chat_textview.update_tag(tag)
+            self.chat_view.update_tag(tag)
 
     def on_close(self, *_args):
-
-        self.frame.notifications.clear("private", self.user)
-        del self.chats.pages[self.user]
-        self.frame.np.privatechats.remove_user(self.user)
-
-        self.chats.remove_page(self.Main)
+        self.core.privatechats.remove_user(self.user)
 
     def on_close_all_tabs(self, *_args):
         self.chats.remove_all_pages()
@@ -454,6 +513,6 @@ class PrivateChat(UserInterface):
 
         # No duplicates
         completion_list = list(set(completion_list))
-        completion_list.sort(key=lambda v: v.lower())
+        completion_list.sort(key=str.lower)
 
         self.chats.completion.set_completion_list(completion_list)
