@@ -1,4 +1,4 @@
-# COPYRIGHT (C) 2020-2022 Nicotine+ Team
+# COPYRIGHT (C) 2020-2022 Nicotine+ Contributors
 # COPYRIGHT (C) 2018 Mutnick <mutnick@techie.com>
 # COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
 # COPYRIGHT (C) 2008-2011 Quinox <quinox@users.sf.net>
@@ -22,18 +22,20 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+from collections import OrderedDict
 from sys import maxsize
 
-from gi.repository import Gio
-from gi.repository import GLib
 from gi.repository import GObject
 from gi.repository import Gtk
 
 from pynicotine.config import config
+from pynicotine.gtkgui.application import GTK_API_VERSION
 from pynicotine.gtkgui.dialogs.fileproperties import FileProperties
 from pynicotine.gtkgui.utils import copy_text
 from pynicotine.gtkgui.widgets.accelerator import Accelerator
 from pynicotine.gtkgui.widgets.popupmenu import PopupMenu
+from pynicotine.gtkgui.widgets.popupmenu import FilePopupMenu
+from pynicotine.gtkgui.widgets.popupmenu import UserPopupMenu
 from pynicotine.gtkgui.widgets.theme import update_widget_visuals
 from pynicotine.gtkgui.widgets.treeview import collapse_treeview
 from pynicotine.gtkgui.widgets.treeview import create_grouping_menu
@@ -41,45 +43,49 @@ from pynicotine.gtkgui.widgets.treeview import initialise_columns
 from pynicotine.gtkgui.widgets.treeview import save_columns
 from pynicotine.gtkgui.widgets.treeview import select_user_row_iter
 from pynicotine.gtkgui.widgets.treeview import show_file_path_tooltip
-from pynicotine.gtkgui.widgets.treeview import verify_grouping_mode
 from pynicotine.gtkgui.widgets.ui import UserInterface
 from pynicotine.transfers import Transfer
 from pynicotine.utils import human_length
 from pynicotine.utils import human_size
 from pynicotine.utils import human_speed
+from pynicotine.utils import humanize
 
 
 class TransferList(UserInterface):
 
-    def __init__(self, frame, transfer_type):
+    path_separator = path_label = retry_label = abort_label = aborted_status = None
+    transfer_page = user_counter = file_counter = expand_button = expand_icon = grouping_button = None
+
+    def __init__(self, frame, core, transfer_type):
 
         super().__init__("ui/" + transfer_type + "s.ui")
-        getattr(frame, transfer_type + "s_content").add(self.Main)
+        (
+            self.clear_all_button,
+            self.container,
+            self.tree_view
+        ) = self.widgets
 
         self.frame = frame
+        self.core = core
         self.type = transfer_type
-        self.page_id = transfer_type + "s"
 
-        self.user_counter = getattr(frame, "%sUsers" % transfer_type.title())
-        self.file_counter = getattr(frame, "%sFiles" % transfer_type.title())
-        grouping_button = getattr(frame, "ToggleTree%ss" % transfer_type.title())
+        if GTK_API_VERSION >= 4:
+            self.clear_all_button.set_has_frame(False)
 
-        if Gtk.get_major_version() == 4:
-            self.ClearTransfers.set_has_frame(False)
+        Accelerator("t", self.tree_view, self.on_abort_transfers_accelerator)
+        Accelerator("r", self.tree_view, self.on_retry_transfers_accelerator)
+        Accelerator("Delete", self.tree_view, self.on_clear_transfers_accelerator)
+        Accelerator("<Alt>Return", self.tree_view, self.on_file_properties_accelerator)
 
-        Accelerator("t", self.Transfers, self.on_abort_transfers_accelerator)
-        Accelerator("r", self.Transfers, self.on_retry_transfers_accelerator)
-        Accelerator("Delete", self.Transfers, self.on_clear_transfers_accelerator)
-        Accelerator("<Alt>Return", self.Transfers, self.on_file_properties_accelerator)
-
-        self.last_ui_update = 0
         self.transfer_list = []
         self.users = {}
         self.paths = {}
-        self.selected_users = []
-        self.selected_transfers = []
         self.tree_users = None
 
+        # Use dict instead of list for faster membership checks
+        self.selected_users = OrderedDict()
+        self.selected_transfers = OrderedDict()
+
         # Status list
         self.statuses = {
             "Queued": _("Queued"),
@@ -87,7 +93,7 @@ class TransferList(UserInterface):
             "Queued (privileged)": _("Queued (privileged)"),
             "Getting status": _("Getting status"),
             "Transferring": _("Transferring"),
-            "Cannot connect": _("Cannot connect"),
+            "Connection timeout": _("Connection timeout"),
             "Pending shutdown.": _("Pending shutdown"),
             "User logged off": _("User logged off"),
             "Disallowed extension": _("Disallowed extension"),  # Sent by Soulseek NS for filtered extensions
@@ -108,7 +114,85 @@ class TransferList(UserInterface):
         }
         self.deprioritized_statuses = ("", "Paused", "Aborted", "Finished", "Filtered")
 
-        self.transfersmodel = Gtk.TreeStore(
+        self.create_model()
+
+        self.column_numbers = list(range(self.transfersmodel.get_n_columns()))
+        self.cols = cols = initialise_columns(
+            frame, transfer_type, self.tree_view,
+            ["user", _("User"), 200, "text", None],
+            ["path", self.path_label, 400, "text", None],
+            ["filename", _("Filename"), 400, "text", None],
+            ["status", _("Status"), 140, "text", None],
+            ["queue_position", _("Queue"), 90, "number", None],
+            ["percent", _("Percent"), 90, "progress", None],
+            ["size", _("Size"), 180, "number", None],
+            ["speed", _("Speed"), 100, "number", None],
+            ["time_elapsed", _("Time Elapsed"), 140, "number", None],
+            ["time_left", _("Time Left"), 140, "number", None],
+        )
+
+        cols["user"].set_sort_column_id(0)
+        cols["path"].set_sort_column_id(1)
+        cols["filename"].set_sort_column_id(2)
+        cols["status"].set_sort_column_id(3)
+        cols["queue_position"].set_sort_column_id(13)
+        cols["percent"].set_sort_column_id(5)
+        cols["size"].set_sort_column_id(10)
+        cols["speed"].set_sort_column_id(12)
+        cols["time_elapsed"].set_sort_column_id(14)
+        cols["time_left"].set_sort_column_id(15)
+
+        menu = create_grouping_menu(
+            frame.window, config.sections["transfers"]["group%ss" % transfer_type], self.on_toggle_tree)
+        self.grouping_button.set_menu_model(menu)
+
+        if GTK_API_VERSION >= 4:
+            self.grouping_button.get_first_child().add_css_class("image-button")
+
+        self.expand_button.connect("toggled", self.on_expand_tree)
+        self.expand_button.set_active(config.sections["transfers"]["%ssexpanded" % transfer_type])
+
+        self.popup_menu_users = UserPopupMenu(frame)
+        self.popup_menu_clear = PopupMenu(frame)
+        self.clear_all_button.set_menu_model(self.popup_menu_clear.model)
+
+        self.popup_menu_copy = PopupMenu(frame)
+        self.popup_menu_copy.add_items(
+            ("#" + _("Copy _File Path"), self.on_copy_file_path),
+            ("#" + _("Copy _URL"), self.on_copy_url),
+            ("#" + _("Copy Folder U_RL"), self.on_copy_dir_url)
+        )
+
+        self.popup_menu = FilePopupMenu(frame, self.tree_view, self.on_popup_menu)
+        self.popup_menu.add_items(
+            ("#" + "selected_files", None),
+            ("", None),
+            ("#" + _("Send to _Player"), self.on_play_files),
+            ("#" + _("_Open in File Manager"), self.on_open_file_manager),
+            ("#" + _("F_ile Properties"), self.on_file_properties),
+            ("", None),
+            ("#" + self.retry_label, self.on_retry_transfer),
+            ("#" + self.abort_label, self.on_abort_transfer),
+            ("#" + _("_Clear"), self.on_clear_transfer),
+            ("", None),
+            ("#" + _("_Browse Folder(s)"), self.on_browse_folder),
+            ("#" + _("_Search"), self.on_file_search),
+            ("", None),
+            (">" + _("Copy"), self.popup_menu_copy),
+            (">" + _("Clear All"), self.popup_menu_clear),
+            (">" + _("User(s)"), self.popup_menu_users)
+        )
+
+        self.file_properties = FileProperties(self.frame, self.core, download_button=False)
+        self.update_visuals()
+
+    def create_model(self):
+        """ Create a tree model based on the grouping mode. Scrolling performance of Gtk.TreeStore
+        is bad with large plain lists, so use Gtk.ListStore in ungrouped mode where no tree structure
+        is necessary. """
+
+        tree_model_class = Gtk.ListStore if self.tree_users == "ungrouped" else Gtk.TreeStore
+        self.transfersmodel = tree_model_class(
             str,                   # (0)  user
             str,                   # (1)  path
             str,                   # (2)  file name
@@ -128,103 +212,29 @@ class TransferList(UserInterface):
             GObject.TYPE_PYOBJECT  # (16) transfer object
         )
 
-        self.column_numbers = list(range(self.transfersmodel.get_n_columns()))
-        self.cols = cols = initialise_columns(
-            frame, transfer_type, self.Transfers,
-            ["user", _("User"), 200, "text", None],
-            ["path", self.path_label, 400, "text", None],
-            ["filename", _("Filename"), 400, "text", None],
-            ["status", _("Status"), 140, "text", None],
-            ["queue_position", _("Queue Position"), 75, "number", None],
-            ["percent", _("Percent"), 70, "progress", None],
-            ["size", _("Size"), 170, "number", None],
-            ["speed", _("Speed"), 90, "number", None],
-            ["time_elapsed", _("Time Elapsed"), 140, "number", None],
-            ["time_left", _("Time Left"), 140, "number", None],
-        )
+        if GTK_API_VERSION == 3:
+            self.transfersmodel.connect("row-changed", self.on_row_changed)
 
-        cols["user"].set_sort_column_id(0)
-        cols["path"].set_sort_column_id(1)
-        cols["filename"].set_sort_column_id(2)
-        cols["status"].set_sort_column_id(3)
-        cols["queue_position"].set_sort_column_id(13)
-        cols["percent"].set_sort_column_id(5)
-        cols["size"].set_sort_column_id(10)
-        cols["speed"].set_sort_column_id(12)
-        cols["time_elapsed"].set_sort_column_id(14)
-        cols["time_left"].set_sort_column_id(15)
-
-        self.Transfers.set_model(self.transfersmodel)
-
-        self.status_page = getattr(frame, "%ss_status_page" % transfer_type)
-        self.expand_button = getattr(frame, "Expand%ss" % transfer_type.title())
-
-        state = GLib.Variant("s", verify_grouping_mode(config.sections["transfers"]["group%ss" % transfer_type]))
-        action = Gio.SimpleAction(name="%sgrouping" % transfer_type, parameter_type=GLib.VariantType("s"), state=state)
-        action.connect("change-state", self.on_toggle_tree)
-        frame.MainWindow.add_action(action)
-        action.change_state(state)
-
-        menu = create_grouping_menu(
-            frame.MainWindow, config.sections["transfers"]["group%ss" % transfer_type], self.on_toggle_tree)
-        grouping_button.set_menu_model(menu)
-
-        self.expand_button.connect("toggled", self.on_expand_tree)
-        self.expand_button.set_active(config.sections["transfers"]["%ssexpanded" % transfer_type])
-
-        self.popup_menu_users = PopupMenu(frame)
-        self.popup_menu_clear = PopupMenu(frame)
-        self.ClearTransfers.set_menu_model(self.popup_menu_clear.model)
-
-        self.popup_menu_copy = PopupMenu(frame)
-        self.popup_menu_copy.add_items(
-            ("#" + _("Copy _File Path"), self.on_copy_file_path),
-            ("#" + _("Copy _URL"), self.on_copy_url),
-            ("#" + _("Copy Folder U_RL"), self.on_copy_dir_url)
-        )
-
-        self.popup_menu = PopupMenu(frame, self.Transfers, self.on_popup_menu)
-        self.popup_menu.add_items(
-            ("#" + "selected_files", None),
-            ("", None),
-            ("#" + _("Send to _Player"), self.on_play_files),
-            ("#" + _("_Open in File Manager"), self.on_open_file_manager),
-            ("#" + _("F_ile Properties"), self.on_file_properties),
-            ("", None),
-            ("#" + _("_Search"), self.on_file_search),
-            ("#" + _("_Browse Folder(s)"), self.on_browse_folder),
-            ("", None),
-            ("#" + self.retry_label, self.on_retry_transfer),
-            ("#" + self.abort_label, self.on_abort_transfer),
-            ("#" + _("_Clear"), self.on_clear_transfer),
-            ("", None),
-            (">" + _("Clear Groups"), self.popup_menu_clear),
-            (">" + _("Copy"), self.popup_menu_copy),
-            (">" + _("User(s)"), self.popup_menu_users)
-        )
-
-        self.update_visuals()
+        if self.tree_users is not None:
+            self.tree_view.set_model(self.transfersmodel)
 
     def init_transfers(self, transfer_list):
         self.transfer_list = transfer_list
-        self.update(forceupdate=True)
+        self.update_model(forceupdate=True)
 
     def server_login(self):
+        # Not needed
         pass
 
     def server_disconnect(self):
+        # Not needed
         pass
 
-    def rebuild_transfers(self):
-        self.clear()
-        self.update()
-
     def save_columns(self):
-        save_columns(self.type, self.Transfers.get_columns())
+        save_columns(self.type, self.tree_view.get_columns())
 
     def update_visuals(self):
-
-        for widget in list(self.__dict__.values()):
+        for widget in self.__dict__.values():
             update_widget_visuals(widget, list_font_target="transfersfont")
 
     def select_transfers(self):
@@ -232,7 +242,7 @@ class TransferList(UserInterface):
         self.selected_transfers.clear()
         self.selected_users.clear()
 
-        model, paths = self.Transfers.get_selection().get_selected_rows()
+        model, paths = self.tree_view.get_selection().get_selected_rows()
 
         for path in paths:
             iterator = model.get_iter(path)
@@ -254,21 +264,14 @@ class TransferList(UserInterface):
         transfer = model.get_value(iterator, 16)
 
         if transfer.filename is not None and transfer not in self.selected_transfers:
-            self.selected_transfers.append(transfer)
+            self.selected_transfers[transfer] = None
 
         if select_user and transfer.user not in self.selected_users:
-            self.selected_users.append(transfer.user)
+            self.selected_users[transfer.user] = None
 
     def new_transfer_notification(self, finished=False):
-        if self.frame.current_page_id != self.page_id:
-            self.frame.request_tab_hilite(self.page_id, mentioned=finished)
-
-    def on_ban(self, *_args):
-
-        self.select_transfers()
-
-        for user in self.selected_users:
-            self.frame.np.network_filter.ban_user(user)
+        if self.frame.current_page_id != self.transfer_page.id:
+            self.frame.notebook.request_tab_hilite(self.transfer_page, mentioned=finished)
 
     def on_file_search(self, *_args):
 
@@ -277,8 +280,8 @@ class TransferList(UserInterface):
         if not transfer:
             return
 
-        self.frame.SearchEntry.set_text(transfer.filename.rsplit("\\", 1)[1])
-        self.frame.change_main_page("search")
+        self.frame.search_entry.set_text(transfer.filename.rsplit("\\", 1)[1])
+        self.frame.change_main_page(self.frame.search_page)
 
     def translate_status(self, status):
 
@@ -290,12 +293,12 @@ class TransferList(UserInterface):
         return status
 
     def update_num_users_files(self):
-        self.user_counter.set_text(str(len(self.users)))
-        self.file_counter.set_text(str(len(self.transfer_list)))
+        self.user_counter.set_text(humanize(len(self.users)))
+        self.file_counter.set_text(humanize(len(self.transfer_list)))
 
-    def update(self, transfer=None, forceupdate=False, update_parent=True):
+    def update_model(self, transfer=None, forceupdate=False, update_parent=True):
 
-        if not forceupdate and self.frame.current_page_id != self.page_id:
+        if not forceupdate and self.frame.current_page_id != self.transfer_page.id:
             # No need to do unnecessary work if transfers are not visible
             return
 
@@ -327,15 +330,14 @@ class TransferList(UserInterface):
                     self.update_parent_row(user_iter, username)
 
             else:
-                for user_path, user_path_iter in list(self.paths.items()):
+                for user_path, user_path_iter in self.paths.copy().items():
                     self.update_parent_row(user_path_iter, user_path, folder=True)
 
-                for username, user_iter in list(self.users.items()):
+                for username, user_iter in self.users.copy().items():
                     self.update_parent_row(user_iter, username)
 
         # Show tab description if necessary
-        self.status_page.set_visible(not self.transfer_list)
-        self.Main.set_visible(self.transfer_list)
+        self.container.get_parent().set_visible(self.transfer_list)
 
     @staticmethod
     def get_hqueue_position(queue_position):
@@ -351,7 +353,7 @@ class TransferList(UserInterface):
 
     @staticmethod
     def get_helapsed(elapsed):
-        return human_length(elapsed) if elapsed >= 1 else ""
+        return human_length(elapsed) if elapsed > 0 else ""
 
     @staticmethod
     def get_hleft(left):
@@ -378,10 +380,9 @@ class TransferList(UserInterface):
     def update_parent_row(self, initer, key, folder=False):
 
         speed = 0.0
-        totalsize = current_bytes = 0
+        total_size = current_bytes = 0
         elapsed = 0
-        left = 0
-        salientstatus = ""
+        salient_status = ""
 
         iterator = self.transfersmodel.iter_children(initer)
 
@@ -396,8 +397,12 @@ class TransferList(UserInterface):
             transfer = self.transfersmodel.get_value(iterator, 16)
             status = transfer.status
 
-            if status == "Transferring" or salientstatus in self.deprioritized_statuses:
-                salientstatus = status
+            if status == "Transferring":
+                speed += transfer.speed or 0
+                salient_status = status
+
+            elif salient_status in self.deprioritized_statuses:
+                salient_status = status
 
             if status == "Filtered":
                 # We don't want to count filtered files when calculating the progress
@@ -405,18 +410,16 @@ class TransferList(UserInterface):
                 continue
 
             elapsed += transfer.time_elapsed or 0
-            left += transfer.time_left or 0
-            totalsize += self.get_size(transfer.size)
+            total_size += self.get_size(transfer.size)
             current_bytes += transfer.current_byte_offset or 0
-            speed += transfer.speed or 0
 
             iterator = self.transfersmodel.iter_next(iterator)
 
         transfer = self.transfersmodel.get_value(initer, 16)
 
-        if transfer.status != salientstatus:
-            self.transfersmodel.set_value(initer, 3, self.translate_status(salientstatus))
-            transfer.status = salientstatus
+        if transfer.status != salient_status:
+            self.transfersmodel.set_value(initer, 3, self.translate_status(salient_status))
+            transfer.status = salient_status
 
         if transfer.speed != speed:
             self.transfersmodel.set_value(initer, 7, self.get_hspeed(speed))
@@ -424,23 +427,23 @@ class TransferList(UserInterface):
             transfer.speed = speed
 
         if transfer.time_elapsed != elapsed:
+            left = (total_size - current_bytes) / speed if speed > 0 else 0
             self.transfersmodel.set_value(initer, 8, self.get_helapsed(elapsed))
             self.transfersmodel.set_value(initer, 9, self.get_hleft(left))
             self.transfersmodel.set_value(initer, 14, elapsed)
             self.transfersmodel.set_value(initer, 15, left)
             transfer.time_elapsed = elapsed
-            transfer.time_left = left
 
         if transfer.current_byte_offset != current_bytes:
-            self.transfersmodel.set_value(initer, 5, self.get_percent(current_bytes, totalsize))
-            self.transfersmodel.set_value(initer, 6, "%s / %s" % (human_size(current_bytes), human_size(totalsize)))
+            self.transfersmodel.set_value(initer, 5, self.get_percent(current_bytes, total_size))
+            self.transfersmodel.set_value(initer, 6, "%s / %s" % (human_size(current_bytes), human_size(total_size)))
             self.transfersmodel.set_value(initer, 11, GObject.Value(GObject.TYPE_UINT64, current_bytes))
             transfer.current_byte_offset = current_bytes
 
-        if transfer.size != totalsize:
-            self.transfersmodel.set_value(initer, 6, "%s / %s" % (human_size(current_bytes), human_size(totalsize)))
-            self.transfersmodel.set_value(initer, 10, GObject.Value(GObject.TYPE_UINT64, totalsize))
-            transfer.size = totalsize
+        if transfer.size != total_size:
+            self.transfersmodel.set_value(initer, 6, "%s / %s" % (human_size(current_bytes), human_size(total_size)))
+            self.transfersmodel.set_value(initer, 10, GObject.Value(GObject.TYPE_UINT64, total_size))
+            transfer.size = total_size
 
     def update_specific(self, transfer=None):
 
@@ -594,50 +597,67 @@ class TransferList(UserInterface):
             if config.sections["ui"]["reverse_file_paths"]:
                 path = self.path_separator.join(reversed(path.split(self.path_separator)))
 
-        iterator = self.transfersmodel.insert_with_values(
-            parent, -1, self.column_numbers,
-            (
-                user,
-                path,
-                shortfn,
-                self.translate_status(status),
-                self.get_hqueue_position(queue_position),
-                self.get_percent(current_byte_offset, size),
-                self.get_hsize(current_byte_offset, size),
-                hspeed,
-                helapsed,
-                self.get_hleft(left),
-                GObject.Value(GObject.TYPE_UINT64, size),
-                GObject.Value(GObject.TYPE_UINT64, current_byte_offset),
-                GObject.Value(GObject.TYPE_UINT64, speed),
-                GObject.Value(GObject.TYPE_UINT, queue_position),
-                elapsed,
-                left,
-                transfer
-            )
+        row = (
+            user,
+            path,
+            shortfn,
+            self.translate_status(status),
+            self.get_hqueue_position(queue_position),
+            self.get_percent(current_byte_offset, size),
+            self.get_hsize(current_byte_offset, size),
+            hspeed,
+            helapsed,
+            self.get_hleft(left),
+            GObject.Value(GObject.TYPE_UINT64, size),
+            GObject.Value(GObject.TYPE_UINT64, current_byte_offset),
+            GObject.Value(GObject.TYPE_UINT64, speed),
+            GObject.Value(GObject.TYPE_UINT, queue_position),
+            elapsed,
+            left,
+            transfer
         )
-        transfer.iterator = iterator
+
+        if parent is None:
+            transfer.iterator = self.transfersmodel.insert_with_valuesv(-1, self.column_numbers, row)
+        else:
+            transfer.iterator = self.transfersmodel.insert_with_values(parent, -1, self.column_numbers, row)
+
         self.update_num_users_files()
 
         if expand_user:
-            self.Transfers.expand_row(self.transfersmodel.get_path(self.users[user]), False)
+            self.tree_view.expand_row(self.transfersmodel.get_path(self.users[user]), False)
 
         if expand_folder:
-            self.Transfers.expand_row(self.transfersmodel.get_path(self.paths[user_path]), False)
+            self.tree_view.expand_row(self.transfersmodel.get_path(self.paths[user_path]), False)
+
+    def clear_model(self):
+
+        self.tree_view.set_model(None)
+
+        self.users.clear()
+        self.paths.clear()
+        self.selected_transfers.clear()
+        self.selected_users.clear()
+        self.transfersmodel.clear()
+
+        self.tree_view.set_model(self.transfersmodel)
+
+        for transfer in self.transfer_list:
+            transfer.iterator = None
 
     def retry_transfers(self):
-        for transfer in self.selected_transfers:
-            getattr(self.frame.np.transfers, "retry_" + self.type)(transfer)
+        # Implemented in subclasses
+        pass
 
     def abort_transfers(self, clear=False):
 
         for transfer in self.selected_transfers:
             if transfer.status != "Finished":
-                self.frame.np.transfers.abort_transfer(transfer, send_fail_message=True)
+                self.core.transfers.abort_transfer(transfer, send_fail_message=True)
 
                 if not clear:
                     transfer.status = self.aborted_status
-                    self.update(transfer)
+                    self.update_model(transfer)
 
             if clear:
                 self.remove_specific(transfer, update_parent=False)
@@ -656,8 +676,8 @@ class TransferList(UserInterface):
             if not self.users[user]:
                 del self.users[user]
 
-        if transfer in self.frame.np.transfers.transfer_request_times:
-            del self.frame.np.transfers.transfer_request_times[transfer]
+        if transfer in self.core.transfers.transfer_request_times:
+            del self.core.transfers.transfer_request_times[transfer]
 
         if not cleartreeviewonly:
             self.transfer_list.remove(transfer)
@@ -669,24 +689,13 @@ class TransferList(UserInterface):
             self.update_parent_rows(transfer)
             self.update_num_users_files()
 
-    def clear_transfers(self, status):
+    def clear_transfers(self, statuses=None):
 
         for transfer in self.transfer_list.copy():
-            if transfer.status in status:
-                self.frame.np.transfers.abort_transfer(transfer, send_fail_message=True)
+            if statuses is None or transfer.status in statuses:
+                self.core.transfers.abort_transfer(transfer, send_fail_message=True)
                 self.remove_specific(transfer)
 
-    def clear(self):
-
-        self.users.clear()
-        self.paths.clear()
-        self.selected_transfers.clear()
-        self.selected_users.clear()
-        self.transfersmodel.clear()
-
-        for transfer in self.transfer_list:
-            transfer.iterator = None
-
     def add_popup_menu_user(self, popup, user):
 
         popup.setup_user_menu(user)
@@ -707,7 +716,7 @@ class TransferList(UserInterface):
         # Multiple users, create submenus for each user
         if len(self.selected_users) > 1:
             for user in self.selected_users:
-                popup = PopupMenu(self.frame)
+                popup = UserPopupMenu(self.frame)
                 self.add_popup_menu_user(popup, user)
                 self.popup_menu_users.add_items((">" + user, popup))
                 self.popup_menu_users.update_model()
@@ -717,20 +726,19 @@ class TransferList(UserInterface):
         user = next(iter(self.selected_users), None)
         self.add_popup_menu_user(self.popup_menu_users, user)
 
-    def on_expand_tree(self, widget):
+    def on_expand_tree(self, *_args):
 
-        expand_button_icon = getattr(self.frame, "Expand%ssImage" % self.type.title())
-        expanded = widget.get_active()
+        expanded = self.expand_button.get_active()
 
         if expanded:
             icon_name = "go-up-symbolic"
-            self.Transfers.expand_all()
+            self.tree_view.expand_all()
 
         else:
             icon_name = "go-down-symbolic"
-            collapse_treeview(self.Transfers, self.tree_users)
+            collapse_treeview(self.tree_view, self.tree_users)
 
-        expand_button_icon.set_property("icon-name", icon_name)
+        self.expand_icon.set_property("icon-name", icon_name)
 
         config.sections["transfers"]["%ssexpanded" % self.type] = expanded
         config.write_configuration()
@@ -739,15 +747,29 @@ class TransferList(UserInterface):
 
         mode = state.get_string()
         active = mode != "ungrouped"
+        grouping_button_style = self.grouping_button.get_parent().get_style_context()
+        popover = self.grouping_button.get_popover()
+
+        if popover is not None:
+            popover.hide()
+
+        # Ensure buttons are flat in libadwaita
+        if active:
+            grouping_button_style.add_class("linked")
+        else:
+            grouping_button_style.remove_class("linked")
 
         config.sections["transfers"]["group%ss" % self.type] = mode
-        self.Transfers.set_show_expanders(active)
+        self.tree_view.set_show_expanders(active)
         self.expand_button.set_visible(active)
 
         self.tree_users = mode
 
+        self.clear_model()
+        self.create_model()
+
         if self.transfer_list:
-            self.rebuild_transfers()
+            self.update_model()
 
         action.set_state(state)
 
@@ -758,12 +780,17 @@ class TransferList(UserInterface):
     def on_popup_menu(self, menu, _widget):
 
         self.select_transfers()
-        num_selected_transfers = len(self.selected_transfers)
-        menu.set_num_selected_files(num_selected_transfers)
+        menu.set_num_selected_files(len(self.selected_transfers))
 
         self.populate_popup_menu_users()
 
-    def on_row_activated(self, _treeview, _path, _column):
+    def on_row_activated(self, _treeview, path, _column):
+
+        if self.tree_view.collapse_row(path):
+            return
+
+        if self.tree_view.expand_row(path, open_all=False):
+            return
 
         self.select_transfers()
         action = config.sections["transfers"]["%s_doubleclick" % self.type]
@@ -789,6 +816,15 @@ class TransferList(UserInterface):
         elif action == 7:  # Browse Folder
             self.on_browse_folder()
 
+    def on_row_changed(self, *_args):
+
+        if self.frame.current_page_id != self.transfer_page.id:
+            return
+
+        # Workaround for GTK 3 issue where GtkTreeView doesn't refresh changed values
+        # if horizontal scrolling is present while fixed-height mode is enabled
+        self.tree_view.queue_draw()
+
     def on_select_user_transfers(self, *args):
 
         if not self.selected_users:
@@ -796,8 +832,8 @@ class TransferList(UserInterface):
 
         selected_user = args[-1]
 
-        sel = self.Transfers.get_selection()
-        fmodel = self.Transfers.get_model()
+        sel = self.tree_view.get_selection()
+        fmodel = self.tree_view.get_model()
         sel.unselect_all()
 
         iterator = fmodel.get_iter_first()
@@ -837,11 +873,14 @@ class TransferList(UserInterface):
     def on_file_properties(self, *_args):
 
         data = []
+        selected_size = 0
 
         for transfer in self.selected_transfers:
             fullname = transfer.filename
             filename = fullname.split("\\")[-1]
             directory = fullname.rsplit("\\", 1)[0]
+            file_size = transfer.size
+            selected_size += file_size
 
             data.append({
                 "user": transfer.user,
@@ -851,13 +890,22 @@ class TransferList(UserInterface):
                 "path": transfer.path,
                 "queue_position": transfer.queue_position,
                 "speed": transfer.speed,
-                "size": transfer.size,
+                "size": file_size,
                 "bitrate": transfer.bitrate,
                 "length": transfer.length
             })
 
         if data:
-            FileProperties(self.frame, data, download_button=False).show()
+            self.file_properties.update_properties(data, total_size=selected_size)
+            self.file_properties.show()
+
+    def on_copy_url(self, *_args):
+        # Implemented in subclasses
+        pass
+
+    def on_copy_dir_url(self, *_args):
+        # Implemented in subclasses
+        pass
 
     def on_copy_file_path(self, *_args):
 
@@ -866,6 +914,18 @@ class TransferList(UserInterface):
         if transfer:
             copy_text(transfer.filename)
 
+    def on_play_files(self, *_args):
+        # Implemented in subclasses
+        pass
+
+    def on_open_file_manager(self, *_args):
+        # Implemented in subclasses
+        pass
+
+    def on_browse_folder(self, *_args):
+        # Implemented in subclasses
+        pass
+
     def on_retry_transfer(self, *_args):
         self.select_transfers()
         self.retry_transfers()
@@ -878,16 +938,13 @@ class TransferList(UserInterface):
         self.select_transfers()
         self.abort_transfers(clear=True)
 
-    def on_clear_response(self, dialog, response_id, data):
-
-        dialog.destroy()
-
+    def on_clear_queued_response(self, _dialog, response_id, _data):
         if response_id == 2:
-            if data == "queued":
-                self.clear_transfers(["Queued"])
+            self.clear_transfers(["Queued"])
 
-            elif data == "all":
-                self.clear()
+    def on_clear_all_response(self, _dialog, response_id, _data):
+        if response_id == 2:
+            self.clear_transfers()
 
     def on_clear_queued(self, *_args):
         self.clear_transfers(["Queued"])
