@@ -1,4 +1,4 @@
-# COPYRIGHT (C) 2020-2022 Nicotine+ Team
+# COPYRIGHT (C) 2020-2022 Nicotine+ Contributors
 # COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
 # COPYRIGHT (C) 2008-2010 Quinox <quinox@users.sf.net>
 # COPYRIGHT (C) 2006-2009 Daelstorm <daelstorm@gmail.com>
@@ -20,55 +20,85 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-import os
 import time
 
+from gi.repository import Gdk
 from gi.repository import GdkPixbuf
+from gi.repository import Gio
 from gi.repository import GLib
 from gi.repository import Gtk
 
 from pynicotine.config import config
 from pynicotine.geoip.geoip import GeoIP
-from pynicotine.gtkgui.widgets.filechooser import save_file
+from pynicotine.gtkgui.application import GTK_API_VERSION
+from pynicotine.gtkgui.widgets.filechooser import FileChooserSave
 from pynicotine.gtkgui.widgets.iconnotebook import IconNotebook
 from pynicotine.gtkgui.widgets.infobar import InfoBar
 from pynicotine.gtkgui.widgets.popupmenu import PopupMenu
+from pynicotine.gtkgui.widgets.popupmenu import UserPopupMenu
 from pynicotine.gtkgui.widgets.textview import TextView
+from pynicotine.gtkgui.widgets.theme import get_flag_icon_name
 from pynicotine.gtkgui.widgets.theme import update_widget_visuals
 from pynicotine.gtkgui.widgets.treeview import initialise_columns
 from pynicotine.gtkgui.widgets.ui import UserInterface
 from pynicotine.logfacility import log
+from pynicotine.slskmessages import UserStatus
 from pynicotine.utils import humanize
 from pynicotine.utils import human_speed
 
 
 class UserInfos(IconNotebook):
 
-    def __init__(self, frame):
+    def __init__(self, frame, core):
 
-        IconNotebook.__init__(self, frame, frame.userinfo_notebook, "userinfo")
-        self.notebook.connect("switch-page", self.on_switch_info_page)
+        super().__init__(
+            frame, core,
+            widget=frame.userinfo_notebook,
+            parent_page=frame.userinfo_page,
+            switch_page_callback=self.on_switch_info_page
+        )
 
     def on_switch_info_page(self, _notebook, page, _page_num):
 
-        if self.frame.current_page_id != self.page_id:
+        if self.frame.current_page_id != self.frame.userinfo_page.id:
             return
 
         for tab in self.pages.values():
-            if tab.Main == page:
-                GLib.idle_add(lambda: tab.descr.grab_focus() == -1)  # pylint:disable=cell-var-from-loop
+            if tab.container == page:
+                GLib.idle_add(lambda tab: tab.description_view.textview.grab_focus() == -1, tab)
                 break
 
+    def on_get_user_info(self, *_args):
+
+        username = self.frame.userinfo_entry.get_text().strip()
+
+        if not username:
+            return
+
+        self.frame.userinfo_entry.set_text("")
+        self.core.userinfo.request_user_info(username)
+
     def show_user(self, user, switch_page=True):
 
         if user not in self.pages:
             self.pages[user] = page = UserInfo(self, user)
-            self.append_page(page.Main, user, page.on_close, user=user)
-            page.set_label(self.get_tab_label_inner(page.Main))
+            self.append_page(page.container, user, page.on_close, user=user)
+            page.set_label(self.get_tab_label_inner(page.container))
 
         if switch_page:
-            self.set_current_page(self.page_num(self.pages[user].Main))
-            self.frame.change_main_page("userinfo")
+            self.set_current_page(self.pages[user].container)
+            self.frame.change_main_page(self.frame.userinfo_page)
+
+    def remove_user(self, user):
+
+        page = self.pages.get(user)
+
+        if page is None:
+            return
+
+        page.clear()
+        self.remove_page(page.container)
+        del self.pages[user]
 
     def show_connection_error(self, user):
         if user in self.pages:
@@ -86,7 +116,7 @@ class UserInfos(IconNotebook):
 
         if msg.user in self.pages:
             page = self.pages[msg.user]
-            self.set_user_status(page.Main, msg.user, msg.status)
+            self.set_user_status(page.container, msg.user, msg.status)
 
     def set_user_country(self, user, country_code):
         if user in self.pages:
@@ -106,7 +136,7 @@ class UserInfos(IconNotebook):
 
     def server_disconnect(self):
         for user, page in self.pages.items():
-            self.set_user_status(page.Main, user, 0)
+            self.set_user_status(page.container, user, UserStatus.OFFLINE)
 
 
 class UserInfo(UserInterface):
@@ -114,29 +144,54 @@ class UserInfo(UserInterface):
     def __init__(self, userinfos, user):
 
         super().__init__("ui/userinfo.ui")
+        (
+            self.container,
+            self.country_icon,
+            self.country_label,
+            self.description_view,
+            self.dislikes_list_view,
+            self.free_upload_slots_label,
+            self.horizontal_paned,
+            self.info_bar,
+            self.likes_list_view,
+            self.picture_container,
+            self.picture_view,
+            self.placeholder_picture,
+            self.progress_bar,
+            self.queued_uploads_label,
+            self.shared_files_label,
+            self.shared_folders_label,
+            self.upload_slots_label,
+            self.upload_speed_label,
+            self.user_label
+        ) = self.widgets
 
         self.userinfos = userinfos
         self.frame = userinfos.frame
+        self.core = userinfos.core
 
-        self.info_bar = InfoBar(self.InfoBar, Gtk.MessageType.INFO)
-        self.descr_textview = TextView(self.descr)
-        self.UserLabel.set_text(user)
+        self.info_bar = InfoBar(self.info_bar, Gtk.MessageType.INFO)
+        self.description_view = TextView(self.description_view)
+        self.user_label.set_text(user)
+
+        if GTK_API_VERSION >= 4:
+            self.country_icon.set_pixel_size(21)
 
-        if Gtk.get_major_version() == 4:
             self.picture = Gtk.Picture(can_shrink=False, halign=Gtk.Align.CENTER, valign=Gtk.Align.CENTER)
-            self.picture_view.set_child(self.picture)
-
             self.scroll_controller = Gtk.EventControllerScroll(flags=Gtk.EventControllerScrollFlags.VERTICAL)
             self.scroll_controller.connect("scroll", self.on_scroll)
             self.picture_view.add_controller(self.scroll_controller)
-
         else:
+            # Setting a pixel size of 21 results in a misaligned country flag
+            self.country_icon.set_pixel_size(0)
+
             self.picture = Gtk.Image(visible=True)
-            self.picture_view.add(self.picture)
             self.picture_view.connect("scroll-event", self.on_scroll_event)
 
+        self.picture_view.set_property("child", self.picture)
+
         self.user = user
-        self.picture_data = None
+        self.picture_data_original = self.picture_data_scaled = None
         self.zoom_factor = 5
         self.actual_zoom = 0
 
@@ -145,29 +200,29 @@ class UserInfo(UserInterface):
 
         self.like_column_numbers = list(range(self.likes_store.get_n_columns()))
         cols = initialise_columns(
-            self.frame, None, self.Likes,
+            self.frame, None, self.likes_list_view,
             ["likes", _("Likes"), 0, "text", None]
         )
         cols["likes"].set_sort_column_id(0)
 
         self.likes_store.set_sort_column_id(0, Gtk.SortType.ASCENDING)
-        self.Likes.set_model(self.likes_store)
+        self.likes_list_view.set_model(self.likes_store)
 
         # Set up dislikes list
-        self.hates_store = Gtk.ListStore(str)
+        self.dislikes_store = Gtk.ListStore(str)
 
-        self.hate_column_numbers = list(range(self.hates_store.get_n_columns()))
+        self.hate_column_numbers = list(range(self.dislikes_store.get_n_columns()))
         cols = initialise_columns(
-            self.frame, None, self.Hates,
+            self.frame, None, self.dislikes_list_view,
             ["dislikes", _("Dislikes"), 0, "text", None]
         )
         cols["dislikes"].set_sort_column_id(0)
 
-        self.hates_store.set_sort_column_id(0, Gtk.SortType.ASCENDING)
-        self.Hates.set_model(self.hates_store)
+        self.dislikes_store.set_sort_column_id(0, Gtk.SortType.ASCENDING)
+        self.dislikes_list_view.set_model(self.dislikes_store)
 
         # Popup menus
-        self.user_popup = popup = PopupMenu(self.frame, None, self.on_tab_popup)
+        self.user_popup = popup = UserPopupMenu(self.frame, None, self.on_tab_popup)
         popup.setup_user_menu(user, page="userinfo")
         popup.add_items(
             ("", None),
@@ -181,10 +236,10 @@ class UserInfo(UserInterface):
                     ("", None),
                     ("#" + _("_Search for Item"), self.on_interest_recommend_search, popup))
 
-        popup = PopupMenu(self.frame, self.Likes, self.on_popup_interest_menu)
+        popup = PopupMenu(self.frame, self.likes_list_view, self.on_popup_interest_menu)
         popup.add_items(*get_interest_items(popup))
 
-        popup = PopupMenu(self.frame, self.Hates, self.on_popup_interest_menu)
+        popup = PopupMenu(self.frame, self.dislikes_list_view, self.on_popup_interest_menu)
         popup.add_items(*get_interest_items(popup))
 
         popup = PopupMenu(self.frame, self.picture_view)
@@ -198,6 +253,13 @@ class UserInfo(UserInterface):
 
         self.update_visuals()
 
+    def clear(self):
+
+        self.description_view.clear()
+        self.likes_store.clear()
+        self.dislikes_store.clear()
+        self.load_picture(None)
+
     def set_label(self, label):
         self.user_popup.set_parent(label)
 
@@ -206,45 +268,48 @@ class UserInfo(UserInterface):
         pass
 
     def update_visuals(self):
-
-        for widget in list(self.__dict__.values()):
+        for widget in self.__dict__.values():
             update_widget_visuals(widget)
 
     """ General """
 
+    def set_pixbuf(self, pixbuf):
+
+        if GTK_API_VERSION >= 4:
+            self.picture.set_pixbuf(pixbuf)
+        else:
+            self.picture.set_from_pixbuf(pixbuf)
+
+        del pixbuf
+
     def load_picture(self, data):
 
         if not data:
-            self.picture_view.hide()
+            if GTK_API_VERSION >= 4:
+                self.picture.set_paintable(None)
+            else:
+                self.picture.clear()
+
+            self.picture_data_original = self.picture_data_scaled = None
+            self.placeholder_picture.show()
             return
 
         try:
-            import gc
-            import tempfile
+            allocation = self.picture_container.get_allocation()
+            max_width = allocation.width - 72
+            max_height = allocation.height - 72
 
-            with tempfile.NamedTemporaryFile() as file_handle:
-                file_handle.write(data)
-                del data
+            # Keep the original picture size for saving to disk
+            data_stream = Gio.MemoryInputStream.new_from_data(data, None)
+            self.picture_data_original = GdkPixbuf.Pixbuf.new_from_stream(data_stream, cancellable=None)
+            picture_width = self.picture_data_original.get_width()
+            picture_height = self.picture_data_original.get_height()
 
-                self.picture_data = GdkPixbuf.Pixbuf.new_from_file(file_handle.name)
-                picture_width = self.picture_data.get_width()
-                picture_height = self.picture_data.get_height()
-
-                allocation = self.placeholder_picture.get_allocation()
-                max_width = allocation.width - 72
-                max_height = allocation.height - 72
-
-                # Resize picture to fit container
-                ratio = min(max_width / picture_width, max_height / picture_height)
-                self.picture_data = self.picture_data.scale_simple(
-                    ratio * picture_width, ratio * picture_height, GdkPixbuf.InterpType.BILINEAR)
-
-                if Gtk.get_major_version() == 4:
-                    self.picture.set_pixbuf(self.picture_data)
-                else:
-                    self.picture.set_from_pixbuf(self.picture_data)
-
-            gc.collect()
+            # Scale picture before displaying
+            ratio = min(max_width / picture_width, max_height / picture_height)
+            self.picture_data_scaled = self.picture_data_original.scale_simple(
+                ratio * picture_width, ratio * picture_height, GdkPixbuf.InterpType.BILINEAR)
+            self.set_pixbuf(self.picture_data_scaled)
 
             self.actual_zoom = 0
             self.picture_view.show()
@@ -256,67 +321,42 @@ class UserInfo(UserInterface):
             })
 
     def make_zoom_normal(self, *_args):
-        self.make_zoom_in(zoom=True)
+        self.actual_zoom = 0
+        self.set_pixbuf(self.picture_data_scaled)
 
-    def make_zoom_in(self, *_args, zoom=None):
+    def make_zoom_in(self, *_args):
 
         def calc_zoom_in(w_h):
             return w_h + w_h * self.actual_zoom / 100 + w_h * self.zoom_factor / 100
 
-        import gc
-
-        if self.picture is None or self.picture_data is None or self.actual_zoom > 100:
+        if self.picture_data_scaled is None or self.actual_zoom >= 100:
             return
 
-        width = self.picture_data.get_width()
-        height = self.picture_data.get_height()
+        self.actual_zoom += self.zoom_factor
+        width = calc_zoom_in(self.picture_data_scaled.get_width())
+        height = calc_zoom_in(self.picture_data_scaled.get_height())
 
-        if zoom:
-            self.actual_zoom = 0
-            picture_zoomed = self.picture_data
-
-        else:
-            self.actual_zoom += self.zoom_factor
-            picture_zoomed = self.picture_data.scale_simple(
-                calc_zoom_in(width), calc_zoom_in(height), GdkPixbuf.InterpType.BILINEAR)
-
-        if Gtk.get_major_version() == 4:
-            self.picture.set_pixbuf(picture_zoomed)
-        else:
-            self.picture.set_from_pixbuf(picture_zoomed)
-
-        del picture_zoomed
-        gc.collect()
+        picture_zoomed = self.picture_data_scaled.scale_simple(width, height, GdkPixbuf.InterpType.NEAREST)
+        self.set_pixbuf(picture_zoomed)
 
     def make_zoom_out(self, *_args):
 
         def calc_zoom_out(w_h):
             return w_h + w_h * self.actual_zoom / 100 - w_h * self.zoom_factor / 100
 
-        import gc
-
-        if self.picture is None or self.picture_data is None:
+        if self.picture_data_scaled is None:
             return
 
-        width = self.picture_data.get_width()
-        height = self.picture_data.get_height()
-
         self.actual_zoom -= self.zoom_factor
+        width = calc_zoom_out(self.picture_data_scaled.get_width())
+        height = calc_zoom_out(self.picture_data_scaled.get_height())
 
-        if calc_zoom_out(width) < 10 or calc_zoom_out(height) < 10:
+        if width < 42 or height < 42:
             self.actual_zoom += self.zoom_factor
             return
 
-        picture_zoomed = self.picture_data.scale_simple(
-            calc_zoom_out(width), calc_zoom_out(height), GdkPixbuf.InterpType.BILINEAR)
-
-        if Gtk.get_major_version() == 4:
-            self.picture.set_pixbuf(picture_zoomed)
-        else:
-            self.picture.set_from_pixbuf(picture_zoomed)
-
-        del picture_zoomed
-        gc.collect()
+        picture_zoomed = self.picture_data_scaled.scale_simple(width, height, GdkPixbuf.InterpType.NEAREST)
+        self.set_pixbuf(picture_zoomed)
 
     def show_connection_error(self):
 
@@ -328,8 +368,8 @@ class UserInfo(UserInterface):
         self.set_finished()
 
     def set_finished(self):
-        self.userinfos.request_tab_hilite(self.Main)
-        self.progressbar.set_fraction(1.0)
+        self.userinfos.request_tab_hilite(self.container)
+        self.progress_bar.set_fraction(1.0)
 
     def message_progress(self, msg):
 
@@ -340,7 +380,7 @@ class UserInfo(UserInterface):
         else:
             fraction = float(msg.position) / msg.total
 
-        self.progressbar.set_fraction(fraction)
+        self.progress_bar.set_fraction(fraction)
 
     """ Network Messages """
 
@@ -350,14 +390,14 @@ class UserInfo(UserInterface):
             return
 
         if msg.descr:
-            self.descr_textview.clear()
-            self.descr_textview.append_line(msg.descr, showstamp=False, scroll=False)
+            self.description_view.clear()
+            self.description_view.append_line(msg.descr)
 
-        self.uploads.set_text(humanize(msg.totalupl))
-        self.queuesize.set_text(humanize(msg.queuesize))
-        self.slotsavail.set_text(_("Yes") if msg.slotsavail else _("No"))
+        self.upload_slots_label.set_text(humanize(msg.totalupl))
+        self.queued_uploads_label.set_text(humanize(msg.queuesize))
+        self.free_upload_slots_label.set_text(_("Yes") if msg.slotsavail else _("No"))
 
-        self.picture_data = None
+        self.picture_data_original = self.picture_data_scaled = None
         self.load_picture(msg.pic)
 
         self.info_bar.set_visible(False)
@@ -366,10 +406,10 @@ class UserInfo(UserInterface):
     def get_user_stats(self, msg):
 
         if msg.avgspeed > 0:
-            self.speed.set_text(human_speed(msg.avgspeed))
+            self.upload_speed_label.set_text(human_speed(msg.avgspeed))
 
-        self.filesshared.set_text(humanize(msg.files))
-        self.dirsshared.set_text(humanize(msg.dirs))
+        self.shared_files_label.set_text(humanize(msg.files))
+        self.shared_folders_label.set_text(humanize(msg.dirs))
 
     def set_user_country(self, country_code):
 
@@ -379,30 +419,32 @@ class UserInfo(UserInterface):
         else:
             country_text = _("Unknown")
 
-        self.country.set_text(country_text)
+        self.country_label.set_text(country_text)
+
+        icon_name = get_flag_icon_name(country_code or "")
+
+        self.country_icon.set_property("icon-name", icon_name)
+        self.country_icon.set_visible(bool(icon_name))
 
     def user_interests(self, msg):
 
         self.likes_store.clear()
-        self.hates_store.clear()
+        self.dislikes_store.clear()
 
         for like in msg.likes:
             self.likes_store.insert_with_valuesv(-1, self.like_column_numbers, [like])
 
         for hate in msg.hates:
-            self.hates_store.insert_with_valuesv(-1, self.hate_column_numbers, [hate])
+            self.dislikes_store.insert_with_valuesv(-1, self.hate_column_numbers, [hate])
 
     """ Callbacks """
 
     def on_tab_popup(self, *_args):
         self.user_popup.toggle_user_items()
 
-    @staticmethod
-    def on_popup_interest_menu(menu, widget):
+    def on_popup_interest_menu(self, menu, widget):
 
-        model, iterator = widget.get_selection().get_selected()
-        item = model.get_value(iterator, 0)
-        menu.set_user(item)
+        item = self.frame.interests.get_selected_item(widget, column=0)
 
         menu.actions[_("I _Like This")].set_state(
             GLib.Variant("b", item in config.sections["interests"]["likes"])
@@ -412,55 +454,54 @@ class UserInfo(UserInterface):
         )
 
     def on_like_recommendation(self, action, state, popup):
-        self.frame.interests.on_like_recommendation(action, state, popup.get_user())
+        item = self.frame.interests.get_selected_item(popup.parent, column=0)
+        self.frame.interests.on_like_recommendation(action, state, item)
 
     def on_dislike_recommendation(self, action, state, popup):
-        self.frame.interests.on_dislike_recommendation(action, state, popup.get_user())
+        item = self.frame.interests.get_selected_item(popup.parent, column=0)
+        self.frame.interests.on_dislike_recommendation(action, state, item)
 
     def on_interest_recommend_search(self, _action, _state, popup):
-        self.frame.interests.recommend_search(popup.get_user())
+        item = self.frame.interests.get_selected_item(popup.parent, column=0)
+        self.frame.interests.recommend_search(item)
 
     def on_send_message(self, *_args):
-        self.frame.np.privatechats.show_user(self.user)
-        self.frame.change_main_page("private")
+        self.core.privatechats.show_user(self.user)
+        self.frame.change_main_page(self.frame.private_page)
 
     def on_show_ip_address(self, *_args):
-        self.frame.np.request_ip_address(self.user)
+        self.core.request_ip_address(self.user)
 
     def on_browse_user(self, *_args):
-        self.frame.np.userbrowse.browse_user(self.user)
+        self.core.userbrowse.browse_user(self.user)
 
     def on_add_to_list(self, *_args):
-        self.frame.np.userlist.add_user(self.user)
+        self.core.userlist.add_user(self.user)
 
     def on_ban_user(self, *_args):
-        self.frame.np.network_filter.ban_user(self.user)
+        self.core.network_filter.ban_user(self.user)
 
     def on_ignore_user(self, *_args):
-        self.frame.np.network_filter.ignore_user(self.user)
+        self.core.network_filter.ignore_user(self.user)
 
-    def on_save_picture_response(self, selected, _data):
-
-        if not os.path.exists(selected):
-            self.picture_data.savev(selected, "jpeg", ["quality"], ["100"])
-            log.add(_("Picture saved to %s"), selected)
-        else:
-            log.add(_("Picture not saved, %s already exists."), selected)
+    def on_save_picture_response(self, file_path, *_args):
+        _success, picture_bytes = self.picture_data_original.save_to_bufferv(
+            type="png", option_keys=[], option_values=[])
+        self.core.userinfo.save_user_picture(file_path, picture_bytes)
 
     def on_save_picture(self, *_args):
 
-        if self.picture is None or self.picture_data is None:
+        if self.picture_data_original is None:
             return
 
-        save_file(
-            parent=self.frame.MainWindow,
+        FileChooserSave(
+            parent=self.frame.window,
             callback=self.on_save_picture_response,
-            initialdir=config.sections["transfers"]["downloaddir"],
-            initialfile="%s %s.jpg" % (self.user, time.strftime("%Y-%m-%d %H_%M_%S")),
-            title=_("Save as…")
-        )
+            initial_folder=config.sections["transfers"]["downloaddir"],
+            initial_file="%s %s.png" % (self.user, time.strftime("%Y-%m-%d %H_%M_%S"))
+        ).show()
 
-    def on_scroll(self, _controller, _scroll_x, scroll_y):
+    def on_scroll(self, _controller=None, _scroll_x=0, scroll_y=0):
 
         if scroll_y < 0:
             self.make_zoom_in()
@@ -471,25 +512,26 @@ class UserInfo(UserInterface):
 
     def on_scroll_event(self, _widget, event):
 
-        if event.get_scroll_deltas().delta_y < 0:
-            self.make_zoom_in()
-        else:
+        if event.direction == Gdk.ScrollDirection.SMOOTH:
+            return self.on_scroll(scroll_y=event.delta_y)
+
+        if event.direction == Gdk.ScrollDirection.DOWN:
             self.make_zoom_out()
 
+        elif event.direction == Gdk.ScrollDirection.UP:
+            self.make_zoom_in()
+
         return True
 
     def on_refresh(self, *_args):
 
         self.info_bar.set_visible(False)
-        self.progressbar.set_fraction(0.0)
+        self.progress_bar.set_fraction(0.0)
 
-        self.frame.np.userinfo.request_user_info(self.user)
+        self.core.userinfo.request_user_info(self.user)
 
     def on_close(self, *_args):
-
-        del self.userinfos.pages[self.user]
-        self.frame.np.userinfo.remove_user(self.user)
-        self.userinfos.remove_page(self.Main)
+        self.core.userinfo.remove_user(self.user)
 
     def on_close_all_tabs(self, *_args):
         self.userinfos.remove_all_pages()
