@@ -1,4 +1,4 @@
-# COPYRIGHT (C) 2020-2022 Nicotine+ Team
+# COPYRIGHT (C) 2020-2022 Nicotine+ Contributors
 # COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
 # COPYRIGHT (C) 2008-2009 Quinox <quinox@users.sf.net>
 # COPYRIGHT (C) 2006-2009 Daelstorm <daelstorm@gmail.com>
@@ -25,11 +25,13 @@ import sys
 from gi.repository import Gdk
 from gi.repository import Gtk
 
-from pynicotine.gtkgui.widgets.dialogs import option_dialog
-from pynicotine.gtkgui.widgets.popupmenu import PopupMenu
-from pynicotine.gtkgui.widgets.theme import get_icon
-from pynicotine.gtkgui.widgets.theme import get_status_icon
 from pynicotine.config import config
+from pynicotine.gtkgui.application import GTK_API_VERSION
+from pynicotine.gtkgui.widgets.dialogs import OptionDialog
+from pynicotine.gtkgui.widgets.popupmenu import PopupMenu
+from pynicotine.gtkgui.widgets.theme import get_status_icon_name
+from pynicotine.gtkgui.widgets.theme import parse_color_string
+from pynicotine.slskmessages import UserStatus
 
 
 """ Icon Notebook """
@@ -46,14 +48,15 @@ class TabLabel(Gtk.Box):
         self.centered = False
         self.gesture_click = None
 
-        if Gtk.get_major_version() == 4:
+        if GTK_API_VERSION >= 4:
             self.eventbox = Gtk.Box()
         else:
             self.eventbox = Gtk.EventBox(visible=True)
+            self.eventbox.add_events(Gdk.EventMask.SCROLL_MASK | Gdk.EventMask.SMOOTH_SCROLL_MASK)
 
         self.box = Gtk.Box(spacing=6, visible=True)
 
-        self.label = Gtk.Label(halign=Gtk.Align.START, hexpand=True, visible=True)
+        self.label = Gtk.Label(halign=Gtk.Align.START, hexpand=True, single_line_mode=True, visible=True)
         self.full_text = full_text
         self.set_text(label)
 
@@ -61,13 +64,8 @@ class TabLabel(Gtk.Box):
         self.close_button_visible = close_button_visible
         self.close_callback = close_callback
 
-        self.start_icon = Gtk.Image()
-        self.start_icon_data = None
-        self.start_icon.hide()
-
-        self.end_icon = Gtk.Image()
-        self.end_icon_data = None
-        self.end_icon.hide()
+        self.start_icon = Gtk.Image(visible=False)
+        self.end_icon = Gtk.Image(visible=False)
 
         self._pack_children()
 
@@ -90,8 +88,9 @@ class TabLabel(Gtk.Box):
         if not self.close_button_visible:
             return
 
-        if Gtk.get_major_version() == 4:
+        if GTK_API_VERSION >= 4:
             self.close_button = Gtk.Button.new_from_icon_name("window-close-symbolic")
+            self.append(self.close_button)  # pylint: disable=no-member
 
             # GTK 4 workaround to prevent notebook tabs from being activated when pressing close button
             gesture_click = Gtk.GestureClick()
@@ -101,12 +100,14 @@ class TabLabel(Gtk.Box):
             self.close_button.add_controller(gesture_click)
 
         else:
-            self.close_button = Gtk.Button.new_from_icon_name("window-close-symbolic", Gtk.IconSize.BUTTON)
+            self.close_button = Gtk.Button.new_from_icon_name("window-close-symbolic",
+                                                              Gtk.IconSize.BUTTON)  # pylint: disable=no-member
+            self.add(self.close_button)  # pylint: disable=no-member
+            self.close_button.add_events(Gdk.EventMask.SCROLL_MASK | Gdk.EventMask.SMOOTH_SCROLL_MASK)
 
         self.close_button.get_style_context().add_class("flat")
         self.close_button.set_tooltip_text(_("Close tab"))
         self.close_button.show()
-        self.add(self.close_button)
 
         if self.close_callback is not None:
             self.close_button.connect("clicked", self.close_callback)
@@ -126,28 +127,37 @@ class TabLabel(Gtk.Box):
             # Left align close button on macOS
             self._add_close_button()
 
-        self.add(self.eventbox)
-        self.eventbox.add(self.box)
+        if GTK_API_VERSION >= 4:
+            self.append(self.eventbox)        # pylint: disable=no-member
+            self.eventbox.append(self.box)    # pylint: disable=no-member
+
+            self.box.append(self.start_icon)  # pylint: disable=no-member
+            self.box.append(self.label)       # pylint: disable=no-member
+            self.box.append(self.end_icon)    # pylint: disable=no-member
+
+        else:
+            self.add(self.eventbox)           # pylint: disable=no-member
+            self.eventbox.add(self.box)       # pylint: disable=no-member
+
+            self.box.add(self.start_icon)     # pylint: disable=no-member
+            self.box.add(self.label)          # pylint: disable=no-member
+            self.box.add(self.end_icon)       # pylint: disable=no-member
+
+        if sys.platform != "darwin":
+            self._add_close_button()
 
         if self.centered:
             self.set_halign(Gtk.Align.CENTER)
         else:
             self.set_halign(Gtk.Align.FILL)
 
-        self.box.add(self.start_icon)
-        self.box.add(self.label)
-        self.box.add(self.end_icon)
-
-        if sys.platform != "darwin":
-            self._add_close_button()
-
     def _set_text_color(self, color):
 
-        color_rgba = Gdk.RGBA()
+        color_hex = parse_color_string(color)
 
-        if color_rgba.parse(color):
+        if color_hex:
             from html import escape
-            self.label.set_markup("<span foreground=\"%s\">%s</span>" % (color, escape(self.text)))
+            self.label.set_markup("<span foreground=\"%s\">%s</span>" % (color_hex, escape(self.text)))
             return
 
         self.label.set_text("%s" % self.text)
@@ -184,16 +194,8 @@ class TabLabel(Gtk.Box):
 
         self._set_text_color(color)
 
-        if self.mentioned:
-            icon_data = get_icon("hilite")
-        else:
-            icon_data = get_icon("hilite3")
-
-        if icon_data is self.end_icon_data:
-            return
-
-        self.end_icon_data = icon_data
-        self.end_icon.set_property("gicon", icon_data)
+        icon_name = "nplus-hilite" if self.mentioned else "nplus-hilite3"
+        self.end_icon.set_property("icon-name", icon_name)
         self.end_icon.show()
 
     def remove_hilite(self):
@@ -203,24 +205,16 @@ class TabLabel(Gtk.Box):
 
         self._set_text_color(config.sections["ui"]["tab_default"])
 
-        self.end_icon_data = None
-        self.end_icon.set_property("gicon", None)
+        self.end_icon.set_property("icon-name", None)
         self.end_icon.hide()
 
     def set_status_icon(self, status):
+        icon_name = get_status_icon_name(status)
+        self.set_start_icon_name(icon_name, visible=config.sections["ui"]["tab_status_icons"])
 
-        icon_data = get_status_icon(status)
-
-        if icon_data is self.start_icon_data:
-            return
-
-        self.start_icon_data = icon_data
-        self.start_icon.set_property("gicon", icon_data)
-        self.start_icon.set_visible(config.sections["ui"]["tab_status_icons"])
-
-    def set_start_icon_name(self, icon_name):
+    def set_start_icon_name(self, icon_name, visible=True):
         self.start_icon.set_property("icon-name", icon_name)
-        self.start_icon.show()
+        self.start_icon.set_visible(visible)
 
     def set_text(self, text):
 
@@ -243,28 +237,37 @@ class IconNotebook:
     - Dropdown menu for unread tabs
     """
 
-    def __init__(self, frame, notebook, page_id):
+    def __init__(self, frame, core, widget, parent_page=None, switch_page_callback=None, reorder_page_callback=None):
 
-        self.notebook = notebook
-        self.notebook.set_show_tabs(False)
-        self.notebook.connect("page-removed", self.on_remove_page)
-        self.notebook.connect("switch-page", self.on_switch_page)
+        self.widget = widget
+        self.widget.connect("page-reordered", self.on_reorder_page)
+        self.widget.connect("page-removed", self.on_remove_page)
+        self.widget.connect("switch-page", self.on_switch_page)
 
         self.frame = frame
-        self.page_id = page_id
+        self.core = core
+        self.parent_page = parent_page
+        self.switch_page_callback = switch_page_callback
+        self.reorder_page_callback = reorder_page_callback
+
         self.unread_button = Gtk.MenuButton(
             tooltip_text=_("Unread Tabs"),
             halign=Gtk.Align.CENTER, valign=Gtk.Align.CENTER, visible=False
         )
         self.pages = {}
+        self.set_show_tabs(False)
 
-        if Gtk.get_major_version() == 4:
-            self.window = self.notebook.get_root()
-            self.unread_button.set_has_frame(False)
-            self.unread_button.set_icon_name("emblem-important-symbolic")
+        if GTK_API_VERSION >= 4:
+            if parent_page is not None:
+                content_box = parent_page.get_first_child()
+                content_box.connect("show", self.on_show_parent_page)
+
+            self.window = self.widget.get_root()
+            self.unread_button.set_has_frame(False)                        # pylint: disable=no-member
+            self.unread_button.set_icon_name("emblem-important-symbolic")  # pylint: disable=no-member
 
             # GTK 4 workaround to prevent notebook tabs from being activated when pressing close button
-            controllers = self.notebook.observe_controllers()
+            controllers = self.widget.observe_controllers()
 
             for num in range(controllers.get_n_items()):
                 item = controllers.get_item(num)
@@ -273,47 +276,73 @@ class IconNotebook:
                     item.set_propagation_phase(Gtk.PropagationPhase.BUBBLE)
                     break
 
+            self.scroll_controller = Gtk.EventControllerScroll(flags=Gtk.EventControllerScrollFlags.BOTH_AXES)
+            self.scroll_controller.connect("scroll", self.on_tab_scroll)
+
+            tab_bar = self.widget.get_first_child()
+            tab_bar.add_controller(self.scroll_controller)
+
         else:
-            self.window = self.notebook.get_toplevel()
-            self.unread_button.set_image(Gtk.Image(icon_name="emblem-important-symbolic"))
+            if parent_page is not None:
+                content_box = parent_page.get_children()[0]
+                content_box.connect("show", self.on_show_parent_page)
+
+            self.window = self.widget.get_toplevel()
+            self.unread_button.set_image(Gtk.Image(icon_name="emblem-important-symbolic"))  # pylint: disable=no-member
+
+            self.widget.add_events(Gdk.EventMask.SCROLL_MASK | Gdk.EventMask.SMOOTH_SCROLL_MASK)
+            self.widget.connect("scroll-event", self.on_tab_scroll_event)
 
         style_context = self.unread_button.get_style_context()
         for style_class in ("circular", "flat"):
             style_context.add_class(style_class)
 
-        self.notebook.set_action_widget(self.unread_button, Gtk.PackType.END)
+        self.widget.set_action_widget(self.unread_button, Gtk.PackType.END)
 
         self.popup_menu_unread = PopupMenu(self.frame, connect_events=False)
         self.unread_button.set_menu_model(self.popup_menu_unread.model)
         self.unread_pages = []
 
-        self.notebook.popup_enable()
+        self.widget.popup_enable()
 
     """ Tabs """
 
     def get_labels(self, page):
-        tab_label = self.notebook.get_tab_label(page)
-        menu_label = self.notebook.get_menu_label(page)
+
+        tab_label = self.widget.get_tab_label(page)
+        menu_label = self.widget.get_menu_label(page)
 
         return tab_label, menu_label
 
     def get_tab_label_inner(self, page):
-        return self.notebook.get_tab_label(page).eventbox
+        tab_label, _menu_label = self.get_labels(page)
+        return tab_label.eventbox
+
+    def set_labels(self, page, tab_label, menu_label):
+        self.widget.set_tab_label(page, tab_label)
+        self.widget.set_menu_label(page, menu_label)
 
     def set_tab_closers(self):
 
-        for i in range(self.notebook.get_n_pages()):
-            page = self.notebook.get_nth_page(i)
+        for i in range(self.get_n_pages()):
+            page = self.get_nth_page(i)
             tab_label, _menu_label = self.get_labels(page)
             tab_label.set_close_button_visibility(config.sections["ui"]["tabclosers"])
 
     def set_tab_text_colors(self):
 
-        for i in range(self.notebook.get_n_pages()):
-            page = self.notebook.get_nth_page(i)
+        for i in range(self.get_n_pages()):
+            page = self.get_nth_page(i)
             tab_label, _menu_label = self.get_labels(page)
             tab_label.set_text(tab_label.get_text())
 
+    def append_page_label(self, page, tab_label, menu_label):
+
+        self.widget.append_page_menu(page, tab_label, menu_label)
+
+        self.set_tab_reorderable(page, True)
+        self.set_show_tabs(True)
+
     def append_page(self, page, text, close_callback=None, full_text=None, user=None):
 
         if full_text is None:
@@ -322,9 +351,9 @@ class IconNotebook:
         label_tab = TabLabel(text, full_text, config.sections["ui"]["tabclosers"], close_callback)
         label_tab.set_tooltip_text(full_text)
 
-        if Gtk.get_major_version() == 4:
+        if GTK_API_VERSION >= 4:
             label_tab.gesture_click = Gtk.GestureClick()
-            label_tab.add_controller(label_tab.gesture_click)
+            label_tab.add_controller(label_tab.gesture_click)  # pylint: disable=no-member
         else:
             label_tab.gesture_click = Gtk.GestureMultiPress(widget=label_tab)
 
@@ -334,97 +363,117 @@ class IconNotebook:
         # menu for all tabs
         label_tab_menu = TabLabel(text)
 
-        self.notebook.append_page_menu(page, label_tab, label_tab_menu)
+        self.append_page_label(page, label_tab, label_tab_menu)
 
         if user is not None:
-            status = 0
-
-            if user in self.frame.np.user_statuses:
-                status = self.frame.np.user_statuses[user] or 0
-
+            status = self.core.user_statuses.get(user, UserStatus.OFFLINE)
             self.set_user_status(page, text, status)
 
-        self.notebook.set_tab_reorderable(page, True)
-        self.notebook.set_show_tabs(True)
-
     def remove_page(self, page):
 
-        self.notebook.remove_page(self.page_num(page))
+        self.widget.remove_page(self.page_num(page))
 
         self.remove_unread_page(page)
 
-        if self.notebook.get_n_pages() == 0:
-            self.notebook.set_show_tabs(False)
+        if self.get_n_pages() == 0:
+            self.set_show_tabs(False)
 
     def remove_all_pages_response(self, dialog, response_id, _data):
 
-        dialog.destroy()
-
         if response_id == 2:
-            for i in reversed(range(self.notebook.get_n_pages())):
-                page = self.notebook.get_nth_page(i)
+            for i in reversed(range(self.get_n_pages())):
+                page = self.get_nth_page(i)
                 tab_label, _menu_label = self.get_labels(page)
                 tab_label.close_callback(dialog)
 
     def remove_all_pages(self):
 
-        option_dialog(
+        OptionDialog(
             parent=self.window,
             title=_('Close All Tabs?'),
             message=_('Do you really want to close all tabs?'),
             callback=self.remove_all_pages_response
-        )
+        ).show()
 
     def get_current_page(self):
-        return self.notebook.get_current_page()
+        return self.get_nth_page(self.widget.get_current_page())
 
-    def set_current_page(self, page_num):
-        return self.notebook.set_current_page(page_num)
+    def set_current_page(self, page):
+        page_num = self.page_num(page)
+        self.widget.set_current_page(page_num)
+
+    def get_current_page_num(self):
+        return self.widget.get_current_page()
+
+    def set_current_page_num(self, page_num):
+        self.widget.set_current_page(page_num)
+
+    def set_show_tabs(self, visible):
+        self.widget.set_show_tabs(visible)
+
+    def set_tab_expand(self, page, expand):
+
+        tab_label, _menu_label = self.get_labels(page)
+
+        if GTK_API_VERSION >= 4:
+            self.widget.get_page(page).set_property("tab-expand", expand)
+        else:
+            self.widget.child_set_property(page, "tab-expand", expand)
+
+        tab_label.set_centered(expand)
+
+    def set_tab_reorderable(self, page, reorderable):
+        self.widget.set_tab_reorderable(page, reorderable)
 
     def set_tab_pos(self, pos):
-        self.notebook.set_tab_pos(pos)
+        self.widget.set_tab_pos(pos)
 
     def get_n_pages(self):
-        return self.notebook.get_n_pages()
+        return self.widget.get_n_pages()
 
     def get_nth_page(self, page_num):
-        return self.notebook.get_nth_page(page_num)
+        return self.widget.get_nth_page(page_num)
 
     def page_num(self, page):
-        return self.notebook.page_num(page)
+        return self.widget.page_num(page)
 
     def next_page(self):
-        return self.notebook.next_page()
+        return self.widget.next_page()
 
     def prev_page(self):
-        return self.notebook.prev_page()
+        return self.widget.prev_page()
+
+    def reorder_child(self, page, order):
+        self.widget.reorder_child(page, order)
 
     """ Tab Highlights """
 
     def request_tab_hilite(self, page, mentioned=False):
 
-        page_active = (self.get_nth_page(self.get_current_page()) == page)
+        if self.parent_page is not None:
+            page_active = (self.get_current_page() == page)
 
-        if self.frame.current_page_id != self.page_id or not page_active:
-            # Highlight top-level tab
-            self.frame.request_tab_hilite(self.page_id, mentioned)
+            if self.frame.current_page_id != self.parent_page.id or not page_active:
+                # Highlight top-level tab
+                self.frame.notebook.request_tab_hilite(self.parent_page, mentioned)
 
-        if page_active:
-            return
+            if page_active:
+                return
+
+            self.append_unread_page(page)
 
         tab_label, menu_label = self.get_labels(page)
         tab_label.request_hilite(mentioned)
         menu_label.request_hilite(mentioned)
 
-        self.append_unread_page(page)
-
     def remove_tab_hilite(self, page):
 
         tab_label, menu_label = self.get_labels(page)
         tab_label.remove_hilite()
         menu_label.remove_hilite()
 
-        self.remove_unread_page(page)
+        if self.parent_page is not None:
+            self.remove_unread_page(page)
 
     def append_unread_page(self, page):
 
@@ -441,12 +490,16 @@ class IconNotebook:
             self.unread_pages.remove(page)
             self.update_unread_pages_menu()
 
-        if not self.unread_pages:
-            self.unread_button.hide()
-            self.frame.remove_tab_hilite(self.page_id)
+        if self.unread_pages:
+            return
+
+        self.unread_button.hide()
+
+        if self.parent_page is not None:
+            self.frame.notebook.remove_tab_hilite(self.parent_page)
 
     def set_unread_page(self, _action, _state, page):
-        self.notebook.set_current_page(self.page_num(page))
+        self.set_current_page(page)
 
     def update_unread_pages_menu(self):
 
@@ -464,13 +517,12 @@ class IconNotebook:
 
     def set_user_status(self, page, user, status):
 
-        if status is None:
-            return
-
-        if status == 1:
+        if status == UserStatus.AWAY:
             status_text = _("Away")
-        elif status == 2:
+
+        elif status == UserStatus.ONLINE:
             status_text = _("Online")
+
         else:
             status_text = _("Offline")
 
@@ -494,12 +546,15 @@ class IconNotebook:
     def on_remove_page(self, _notebook, new_page, _page_num):
         self.remove_unread_page(new_page)
 
-    def on_switch_page(self, _notebook, new_page, _page_num):
+    def on_switch_page(self, _notebook, new_page, page_num):
+
+        if self.switch_page_callback is not None:
+            self.switch_page_callback(self, new_page, page_num)
 
         # Hide container widget on previous page for a performance boost
-        current_page = self.get_nth_page(self.get_current_page())
+        current_page = self.get_current_page()
 
-        if Gtk.get_major_version() == 4:
+        if GTK_API_VERSION >= 4:
             current_page.get_first_child().hide()
             new_page.get_first_child().show()
         else:
@@ -507,7 +562,51 @@ class IconNotebook:
             new_page.get_children()[0].show()
 
         # Dismiss tab highlight
-        self.remove_tab_hilite(new_page)
+        if self.parent_page is not None:
+            self.remove_tab_hilite(new_page)
+
+    def on_reorder_page(self, _notebook, page, page_num):
+        if self.reorder_page_callback is not None:
+            self.reorder_page_callback(self, page, page_num)
+
+    def on_show_parent_page(self, _widget):
+
+        curr_page = self.get_current_page()
+        curr_page_num = self.get_current_page_num()
+
+        if curr_page_num >= 0:
+            self.widget.emit("switch-page", curr_page, curr_page_num)
+
+    def on_tab_scroll_event(self, _widget, event):
+
+        current_page = self.get_current_page()
+
+        if not current_page:
+            return False
+
+        if Gtk.get_event_widget(event).is_ancestor(current_page):
+            return False
+
+        if event.direction == Gdk.ScrollDirection.SMOOTH:
+            return self.on_tab_scroll(scroll_x=event.delta_x, scroll_y=event.delta_y)
+
+        if event.direction in (Gdk.ScrollDirection.RIGHT, Gdk.ScrollDirection.DOWN):
+            self.next_page()
+
+        elif event.direction in (Gdk.ScrollDirection.LEFT, Gdk.ScrollDirection.UP):
+            self.prev_page()
+
+        return True
+
+    def on_tab_scroll(self, _controller=None, scroll_x=0, scroll_y=0):
+
+        if scroll_x > 0 or scroll_y > 0:
+            self.next_page()
+
+        elif scroll_x < 0 or scroll_y < 0:
+            self.prev_page()
+
+        return True
 
     def on_tab_popup(self, widget, page):
         # Dummy implementation
